import glob
import os
import numpy as np
from scipy.integrate import solve_ivp
import scipy
from pytest import raises

from pydmd.mrcosts import mrCOSTS


def build_multiscale_process():
    """
    Simulates a system with two oscillators with occasionally overlapping
    frequencies (adapted from Dylewsky et al., 2019) with a transient
    coherent wave packet constructive interfering with the system.

    Oscillator #1: FitzHugh-Nagumo Model
    Oscillator #2: Unforced Duffing Oscillator

    Data are sub-selected to reduce the computational load from minutes to
    10s of seconds. See costs-tutorial_toy-data for details.
    """

    def rhs_FNM(t, x, tau, a, b, Iext):
        """FitzHugh-Nagumo Model

        See costs-tutorial_toy-data for details.
        """
        v = x[0]
        w = x[1]
        vdot = v - (v**3) / 3 - w + Iext
        wdot = (1 / tau) * (v + a - b * w)
        dx = np.array([vdot, wdot])

        return dx

    def rhs_UFD(t, y, eta, epsilon, tau):
        """Unforced Duffing Oscillator

        See costs-tutorial_toy-data for details.
        """
        p = y[0]
        q = y[1]
        pdot = q
        qdot = (1 / tau) * (-2 * eta * q - p - epsilon * p**3)
        dy = np.array([pdot, qdot])

        return dy

    T = 64

    x0 = np.array([-1.110, -0.125])
    tau1 = 2
    a = 0.7
    b = 0.8
    Iext = 0.65

    y0 = np.array([0, 1])
    eta = 0
    epsilon = 1
    tau2 = 0.2

    # RK4 integration of the mixed system
    dt = 0.0001 * 8
    t_solution = np.arange(0, T, dt)

    # Solve the FitzHugh-Nagumo Model
    solution_fn = solve_ivp(
        rhs_FNM, [0, T], x0, t_eval=t_solution, args=(tau1, a, b, Iext)
    )

    # Solve the Unforced Duffing Oscillator Model
    solution_ufd = solve_ivp(
        rhs_UFD, [0, T], y0, t_eval=t_solution, args=(eta, epsilon, tau2)
    )

    seed = 1
    num_space_dims = 4

    uv_tiled = np.hstack(
        [
            np.tile(solution_fn.y.T, num_space_dims),
            np.tile(solution_ufd.y.T, num_space_dims),
        ]
    )

    # Subsample after solving the pdes
    substep = 100
    uv_tiled = uv_tiled[0::substep, :]
    t_solution = t_solution[0::substep]

    # Dimension of space to map into
    n_space_dims = np.shape(uv_tiled)[1]
    n_time = np.shape(uv_tiled)[0]

    # Orthonormalized linear mixing matrix
    Q = scipy.stats.ortho_group.rvs(n_space_dims, random_state=seed)
    Q = Q[0:n_space_dims, :]
    x = uv_tiled @ Q

    # COSTS expects time by space, so we transpose x.
    data_original = x.T

    # For the scale separation we want to compare to the actual slow and fast
    # components.
    slow_modes = (
        uv_tiled[:, 0 : n_space_dims // 2] @ Q[0 : n_space_dims // 2, :]
    )
    fast_modes = uv_tiled[:, n_space_dims // 2 :] @ Q[n_space_dims // 2 :, :]

    # Add a transient wave packet
    recon_filter_sd = len(t_solution) * 0.25

    recon_filter = np.exp(
        -((np.arange(n_time) - (n_time + 1) / 2) ** 2) / recon_filter_sd**2
    )
    recon_filter[recon_filter < 0.0001] = 0
    f_transient = 10
    x_transient = (
        0.5
        * np.sin(f_transient * t_solution.flatten())
        * np.sin(0.25 * t_solution.flatten())
        * recon_filter
    )

    # Add the transient feature to the data
    data_transient = data_original + np.atleast_2d(x_transient)

    return (
        t_solution,
        data_transient,
        slow_modes.T,
        fast_modes.T,
        x_transient.T,
    )


# Simulate data.
(
    time,
    data,
    low_frequency,
    high_frequency,
    transient,
) = build_multiscale_process()
# Define the true eigenvalues of the system.
expected_frequency_bands = np.array((0.4, 1.0))
expected_n_components = 2

# Define the expected error in the reconstructions.
expected_global_error = 0.053
expected_lf_error = 0.10
expected_hf_error = 0.16
expected_transient_error = 0.3

# Fit mrCOSTS for testing
window_lengths = [15, 60]
step_sizes = [1, 12]
svd_ranks = [4] * len(window_lengths)
suppress_growth = True
transform_method = "square_frequencies"
n_components_array = [2] * len(window_lengths)
global_svd_array = [True] * len(window_lengths)

mrc = mrCOSTS(
    svd_rank_array=svd_ranks,
    window_length_array=window_lengths,
    step_size_array=step_sizes,
    global_svd_array=global_svd_array,
    transform_method=transform_method,
    n_components_array=n_components_array,
)
mrc.fit(data, np.atleast_2d(time))
mrc.multi_res_interp()

# Global clustering
n_components_range = np.arange(2, 8)
scores, n_optimal = mrc.global_cluster_hyperparameter_sweep(
    n_components_range, transform_method="log10", verbose=False
)

cluster_centroids, omega_classes, omega_array = mrc.global_cluster_omega(
    transform_method="log10"
)


def test_frequency_band_centroids():
    """
    Tests that the identified frequency bands are near the expected values.
    """

    np.testing.assert_allclose(
        mrc.cluster_centroids, expected_frequency_bands, atol=0.1
    )
    np.testing.assert_equal(np.unique(mrc.omega_classes_interpolated), [0, 1])
    assert mrc._n_components_global == 2

    assert len(mrc.ragged_omega_classes) == len(window_lengths)
    om_cl = mrc.ragged_omega_classes
    np.testing.assert_equal(np.unique(om_cl[0]), [-1, 1])
    np.testing.assert_equal(np.unique(om_cl[1]), [-1, 0])

    assert mrc.transform_method == transform_method

    # Verify these fields are empty prior to clustering.
    mrc_no_clustering = mrCOSTS()
    with raises(ValueError):
        mrc_no_clustering.n_components_global
    with raises(ValueError):
        mrc_no_clustering.cluster_centroids
    assert mrc_no_clustering.transform_method is None


def test_reconstructions():
    """
    Tests the accuracy of the reconstructed data.
    """
    xr_sep = mrc.global_scale_reconstruction()
    xr_hf = xr_sep.sum(axis=0)[0, :, :].squeeze()
    xr_transient = xr_sep.sum(axis=0)[1, :, :].squeeze()
    xr_lf = mrc.get_background()
    xr_global = mrc.global_reconstruction()

    mrd = mrc.costs_array[0]
    re_global = mrd.relative_error(xr_global, data)
    re_lf = mrd.relative_error(xr_lf, low_frequency)
    re_hf = mrd.relative_error(xr_hf, high_frequency)

    n_data_vars, n_time_steps = np.shape(data)

    error_array = np.zeros(n_data_vars)
    for n in range(n_data_vars):
        error_array[n] = mrc.costs_array[0].relative_error(
            xr_transient[n, :], xr_transient
        )
    re_transient = error_array.mean()

    np.testing.assert_allclose(re_global, expected_global_error, atol=0.01)
    np.testing.assert_allclose(re_lf, expected_lf_error, atol=0.01)
    np.testing.assert_allclose(re_hf, expected_hf_error, atol=0.01)
    # There is a strong random component to this test, necessitating very loose
    # tolerances.
    np.testing.assert_allclose(
        re_transient, expected_transient_error, atol=0.15
    )


def test_omega_transforms():
    """
    Tests that the COSTS module correctly transforms the eigenvalues yielding...
    - absolute values for "absolute"
    - squared eigenvalues for "square_frequencies"
    - 1 / absolute eigenvalues for "period"
    - log10 of the eignevalues for "log10"
    Eigenvalue constraint combinations are also tested.
    """

    list_transform_methods = [
        "absolute",
        "square_frequencies",
        "period",
        "log10",
    ]

    omega_to_transform = mrc.ragged_omega_array[1]

    for method in list_transform_methods:
        mrc_transformer = mrCOSTS()

        omega_transformed = mrc_transformer.transform_omega(
            omega_to_transform, transform_method=method
        )
        if method != "log10":
            assert np.all(omega_transformed >= 0.0)
        elif method == "log10":
            assert np.allclose(
                np.log10(np.max(np.abs(omega_to_transform.imag))),
                np.max(omega_transformed),
            )
        assert np.all(np.isfinite(omega_transformed))

    with raises(ValueError):
        mrc_transformer = mrCOSTS()
        mrc_transformer.transform_omega(
            omega_to_transform, transform_method="bad"
        )


def test_netcdf(tmp_path):
    """
    Test the round trip conversion of the mrCOSTS object to file in
    netcdf format and back to mrCOSTS.
    """
    # Move the I/O tests to the temporary test directory.
    mrc.to_netcdf("tests", filepath=tmp_path)
    file_list = glob.glob(os.path.join(tmp_path, "*tests*.nc"))
    mrc_from_file = mrCOSTS()
    mrc_from_file.from_netcdf(file_list)

    for n in range(mrc.n_decompositions):
        assert np.allclose(
            mrc.ragged_modes_array[n], mrc_from_file.ragged_modes_array[n]
        )
    assert np.allclose(mrc.cluster_centroids, mrc.cluster_centroids)


def test_plot_local_reconstructions():
    """
    Test that the local reconstruction plot works and raises the appropriate
    errors when trying to plot the first decomposition level without data.
    """
    mrc.plot_local_reconstructions(0, data=data)

    with raises(ValueError):
        mrc.plot_local_reconstructions(0)

    with raises(ValueError):
        mrc.plot_local_reconstructions(0, data=data.T)


def test_plot_local_error():
    """
    Test that the local error plot works and raises the appropriate
    errors when trying to plot the first decomposition level without data.
    """
    mrc.plot_local_error(0, data=data)

    with raises(ValueError):
        mrc.plot_local_error(0)

    with raises(ValueError):
        mrc.plot_local_error(0, data=data.T)


def test_plot_local_scale_separation():
    """
    Test that the local scale separation plot works and raises the appropriate
    errors when trying to plot the first decomposition level without data.
    """
    _ = mrc.plot_local_scale_separation(0, data=data)

    with raises(ValueError):
        _ = mrc.plot_local_scale_separation(0)

    with raises(ValueError):
        _ = mrc.plot_local_scale_separation(0, data=data.T)


def test_plot_local_time_series():
    """
    Test that the local time series of a point plot works and raises the
    appropriate errors when trying to plot the first decomposition level
    without data.
    """
    _ = mrc.plot_local_time_series(0, 0, data=data)

    with raises(ValueError):
        _ = mrc.plot_local_time_series(0, 0)

    with raises(ValueError):
        _ = mrc.plot_local_time_series(0, 0, data=data.T)



"""
PyDMD preprocessing init.
"""

from .hankel import hankel_preprocessing
from .pre_post_processing import PrePostProcessingDMD
from .randomized import randomized_preprocessing
from .svd_projection import svd_projection_preprocessing
from .zero_mean import zero_mean_preprocessing



import numpy as np
from scipy.integrate import solve_ivp
import scipy
from pytest import raises

from pydmd.costs import COSTS


def overlapping_oscillators():
    """
    Simulates a system with two oscillators with ocasionally overlapping
    frequencies. This example was adapted from Dylewsky et al., 2019.

    Oscillator #1: FitzHugh-Nagumo Model
    Oscillator #2: Unforced Duffing Oscillator
    """

    def rhs_FNM(t, x, tau, a, b, Iext):
        """FitzHugh-Nagumo Model

        See costs-tutorial_toy-data for details.
        """
        v = x[0]
        w = x[1]
        vdot = v - (v**3) / 3 - w + Iext
        wdot = (1 / tau) * (v + a - b * w)
        dx = np.array([vdot, wdot])

        return dx

    def rhs_UFD(t, y, eta, epsilon, tau):
        """Unforced Duffing Oscillator

        See costs-tutorial_toy-data for details.
        """
        p = y[0]
        q = y[1]
        pdot = q
        qdot = (1 / tau) * (-2 * eta * q - p - epsilon * p**3)
        dy = np.array([pdot, qdot])

        return dy

    T = 64

    x0 = np.array([-1.110, -0.125])
    tau1 = 2
    a = 0.7
    b = 0.8
    Iext = 0.65

    y0 = np.array([0, 1])
    eta = 0
    epsilon = 1
    tau2 = 0.2

    # RK4 integration of the mixed system
    dt = 0.0001 * 8
    t_solution = np.arange(0, T, dt)

    # Solve the FitzHugh-Nagumo Model
    solution_fn = solve_ivp(
        rhs_FNM, [0, T], x0, t_eval=t_solution, args=(tau1, a, b, Iext)
    )

    # Solve the Unforced Duffing Oscillator Model
    solution_ufd = solve_ivp(
        rhs_UFD, [0, T], y0, t_eval=t_solution, args=(eta, epsilon, tau2)
    )

    seed = 1
    num_space_dims = 4

    uv_tiled = np.hstack(
        [
            np.tile(solution_fn.y.T, num_space_dims),
            np.tile(solution_ufd.y.T, num_space_dims),
        ]
    )

    # Subsample after solving the pdes
    substep = 100
    uv_tiled = uv_tiled[0::substep, :]
    t_solution = t_solution[0::substep]

    # Dimension of space to map into
    n_space_dims = np.shape(uv_tiled)[1]

    # Orthonormalized linear mixing matrix
    Q = scipy.stats.ortho_group.rvs(n_space_dims, random_state=seed)
    Q = Q[0:n_space_dims, :]
    x = uv_tiled @ Q

    # COSTS expects time by space, so we transpose x.
    solution = x.T

    # For the scale separation we want to compare to the actual slow and fast
    # components.
    slow_modes = (
        uv_tiled[:, 0 : n_space_dims // 2] @ Q[0 : n_space_dims // 2, :]
    )
    fast_modes = uv_tiled[:, n_space_dims // 2 :] @ Q[n_space_dims // 2 :, :]

    return t_solution, solution, slow_modes.T, fast_modes.T


# Simulate data.
time, data, low_frequency, high_frequency = overlapping_oscillators()
# Define the true eigenvalues of the system.
expected_frequency_bands = np.array((1.1, 2.5))
# Define the expected error in the reconstructions.
expected_global_error = 0.045
expected_lf_error = 0.11
expected_hf_error = 0.17
expected_n_slides = 63

# Fit COSTS once and then test.
window = 60
step = 12
rank = 4
transform_method = "absolute"
pydmd_kwargs = {"eig_constraints": {"conjugate_pairs", "stable"}}
mrd = COSTS(
    svd_rank=rank,
    global_svd=True,
    pydmd_kwargs=pydmd_kwargs,
)
mrd.fit(data, np.atleast_2d(time), window, step, verbose=False)
# Force the clustering to use two components due to the nature of the toy data.
mrd.cluster_omega(n_components=2, transform_method=transform_method)


def test_construction():
    """
    Test basic properties of building the windowed decomposition.
    """

    n_space_dims = np.shape(data)[0]

    assert mrd.n_slides == expected_n_slides
    assert mrd.window_length == window
    assert np.shape(mrd.time_array) == (mrd.n_slides, mrd.window_length)
    assert np.shape(mrd.modes_array) == (
        mrd.n_slides,
        n_space_dims,
        mrd.svd_rank,
    )
    assert np.shape(mrd.amplitudes_array) == (
        mrd.n_slides,
        mrd.svd_rank,
    )
    assert mrd.global_svd is True
    assert mrd.step_size == step


def test_bad_construction():
    """Test bad fit and construction keywords and parameters."""
    mrd_alternative = COSTS()
    with raises(ValueError):
        mrd_alternative.fit(
            data, np.atleast_2d(time), len(time) + 1, step, verbose=False
        )


def test_integer_window_construction():
    """Force COSTS to build an integer number of windows."""
    assert mrd._build_windows(data, window, step, integer_windows=True) == 61


def test_frequency_band_centroids():
    """
    Tests that the identified frequency bands are near the expected values.
    """

    np.testing.assert_allclose(
        mrd.cluster_centroids, expected_frequency_bands, atol=0.1
    )
    np.testing.assert_equal(np.unique(mrd.omega_classes), [0, 1])
    assert mrd.n_components == 2

    mrd_no_clustering = COSTS()
    assert mrd_no_clustering.n_components is None
    assert mrd_no_clustering.cluster_centroids is None


def test_reconstruction():
    """
    Tests the accuracy of the reconstructed data.
    """
    global_reconstruction = mrd.global_reconstruction()
    re_global = mrd.relative_error(global_reconstruction.real, data)
    np.testing.assert_allclose(re_global, expected_global_error, atol=0.01)


def test_scale_separation():
    """
    Tests the accuracy of the scale separation.
    """
    costs_lf, costs_hf = mrd.scale_separation()
    re_lf = mrd.relative_error(costs_lf.real, low_frequency)
    np.testing.assert_allclose(re_lf, expected_lf_error, atol=0.02)
    re_hf = mrd.relative_error(costs_hf.real, high_frequency)
    np.testing.assert_allclose(re_hf, expected_hf_error, atol=0.02)


def test_omega_transforms():
    """
    Tests that the COSTS module correctly transforms the eigenvalues yielding...
    - absolute values for "absolute"
    - squared eigenvalues for "square_frequencies"
    - 1 / absolute eigenvalues for "period"
    - log10 of the eignevalues for "log10"
    - a bad transformation which shouldn't work.
    """

    list_transform_methods = [
        "absolute",
        "square_frequencies",
        "period",
        "log10",
    ]

    for method in list_transform_methods:
        mrd_transform = COSTS(
            svd_rank=rank,
            global_svd=True,
            pydmd_kwargs=pydmd_kwargs,
        )

        omega_transform = mrd_transform.transform_omega(
            mrd.omega_array, transform_method=method
        )
        if method != "log10":
            assert np.all(omega_transform >= 0.0)
        elif method == "log10":
            assert np.allclose(
                np.log10(np.max(np.abs(mrd.omega_array.imag))),
                np.max(omega_transform),
            )
        assert np.all(np.isfinite(omega_transform))
        assert mrd_transform._omega_label is not None
        assert mrd_transform._hist_kwargs is not None

    with raises(ValueError):
        mrd_transform = COSTS(
            svd_rank=rank,
            global_svd=True,
            pydmd_kwargs=pydmd_kwargs,
        )

        mrd_transform.transform_omega(mrd.omega_array, transform_method="bad")


def test_to_xarray():
    """Tests the round trip conversion to and from xarray."""
    ds = mrd.to_xarray()
    mrd_convert = mrd.from_xarray(ds)

    assert np.allclose(mrd.omega_array, mrd_convert.omega_array)
    assert np.allclose(mrd.modes_array, mrd_convert.modes_array)
    assert np.allclose(mrd.cluster_centroids, mrd_convert.cluster_centroids)

    # The round trip of the pydmd_kwargs is sensitive to python and numpy
    # version.
    assert np.allclose(
        mrd._pydmd_kwargs["proj_basis"], mrd_convert._pydmd_kwargs["proj_basis"]
    )
    for kw in mrd._pydmd_kwargs:
        if not kw == "proj_basis":
            assert mrd._pydmd_kwargs[kw] == mrd_convert._pydmd_kwargs[kw]


def test_plotters():
    """Determine no errors are triggered when plotting."""
    mrd.plot_reconstructions(data)
    mrd.plot_scale_separation(data)
    mrd.plot_error(data)
    mrd.plot_time_series(1, data)
    mrd.plot_omega_histogram()
    mrd.plot_omega_time_series()



"""
Derived module from dmdbase.py for kernelized extended dmd.

References:
- M. O. Williams, C. W. Rowley, and I. G. Kevrekidis,
A kernel-based method for data-driven koopman spectral analysis,
J. Comput. Dynam., 2 (2015), pp. 247-265
- M. O. Williams, I. G. Kevrekidis, and C. W. Rowley,
A data-driven approximation of the koopman operator: extending
dynamic mode decomposition, J. Nonlinear Sci., 25 (2015), pp. 1307-1346
"""

import numpy as np
from sklearn.metrics.pairwise import pairwise_kernels

from .dmd import DMD
from .dmdoperator import DMDOperator


class EDMDOperator(DMDOperator):
    """
    DMD operator for kernelized extended DMD.

    :param svd_rank: the rank for the truncation; If positive integer, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param kernel_metric: the kernel function to apply. Supported kernel
        metrics include `"additive_chi2"`, `"chi2"`, `"linear"`, `"poly"`,
        `"polynomial"`, `"rbf"`, `"laplacian"`, `"sigmoid"`, and `"cosine"`.
    :type kernel_metric: str
    :param kernel_params: additional parameters for the
        `sklearn.metrics.pairwise_kernels` function, including
        kernel-specific function parameters.
    :type kernel_params: dict
    """

    def __init__(self, svd_rank, kernel_metric, kernel_params):
        super().__init__(
            svd_rank=svd_rank,
            exact=True,
            forward_backward=False,
            rescale_mode=None,
            sorted_eigs=False,
            tikhonov_regularization=None,
        )
        self._kernel_metric = kernel_metric
        self._kernel_params = kernel_params
        self._eigenvalues = None
        self._Lambda = None
        self._modes = None
        self._svd_vals = None
        self._eigenvectors = None

    @property
    def svd_vals(self):
        """
        :return: the singular values of the extended feature matrix.
        :rtype: numpy.ndarray
        """
        return self._svd_vals

    @property
    def as_numpy_array(self):
        """
        EDMD does not explicitly compute the forward operator or its reduction.
        Doing so poses risks of prohibitively-large computations...
        """
        raise ValueError("Atilde is not computed explicitly during EDMD.")

    def compute_operator(self, X, Y):
        """
        Compute and store the kernelized EDMD diagnostics.

        :param X: matrix containing the right-hand side snapshots by column.
        :type X: numpy.ndarray
        :param Y: matrix containing the left-hand side snapshots by column.
        :type Y: numpy.ndarray
        :return: the (truncated) left-singular vectors matrix of the extended
            feature matrix, the (truncated) singular values matrix of the
            extended feature matrix, and the eigenvectors of the tranformed
            forward operator `K_hat`.
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray
        """
        # Compute kernel matrices using given snapshots and kernel function.
        G_hat = pairwise_kernels(
            X.T, X.T, metric=self._kernel_metric, **self._kernel_params
        )
        A_hat = pairwise_kernels(
            Y.T, X.T, metric=self._kernel_metric, **self._kernel_params
        )

        # Use the Gramian to obtain the singular vectors Q and singular
        # values Sigma of the feature embedding matrix.
        Q, Sigma = self._compute_feature_matrix_svd(G_hat)

        # Construct the tranformed forward operator matrix K_hat.
        K_hat = np.linalg.multi_dot(
            [np.linalg.pinv(Sigma), Q.T, A_hat, Q, np.linalg.pinv(Sigma)]
        )

        # Obtain the eigenvalues and eigenvectors of K_hat.
        # Note: The nonzero eigenvalues of K_hat are the DMD eigenvalues.
        # DMD modes are derived via the left and right eigenvectors of K_hat.
        eigenvalues, V_hat = np.linalg.eig(K_hat)
        modes = np.linalg.multi_dot(
            [X, Q, np.linalg.pinv(Sigma), np.linalg.inv(V_hat).T]
        )

        # Discard eigenvalue, mode pairs that correspond with zero eigenvalues.
        nonzero_inds = np.abs(eigenvalues) > 1e-16
        eigenvalues = eigenvalues[nonzero_inds]
        modes = modes[:, nonzero_inds]

        # Set the DMD eigenvalues and modes.
        self._eigenvalues = eigenvalues
        self._Lambda = eigenvalues
        self._modes = modes

        # Store Sigma and V_hat for eigenfunction computations.
        self._svd_vals = np.diag(Sigma)
        self._eigenvectors = V_hat

        return Q, Sigma, V_hat

    def _compute_feature_matrix_svd(self, Gramian):
        """
        Helper function that computes the eigendecomposition of the Gramian in
        order to obtain the singular vectors Q and singular values Sigma of the
        corresponding feature embedding matrix.

        :param numpy.ndarray Gramian: the Gramian matrix.
        :return: the (truncated) left-singular vectors matrix of the extended
            feature matrix and the (truncated) singular values matrix of the
            extended feature matrix.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        G_eigenvalues, G_eigenvectors = np.linalg.eig(Gramian)

        # Note: The eigenvalues of the Gramian are the singular values of the
        # feature embedding matrix, squared. Hence in theory, the eigenvalues
        # should be real and non-negative, so we discard imaginary components
        # and flip the sign of negative eigenvalues.
        G_eigenvalues = G_eigenvalues.real
        negative_inds = G_eigenvalues < 0
        G_eigenvectors[:, negative_inds] *= -1
        G_eigenvalues[negative_inds] *= -1
        s_vals = np.sqrt(G_eigenvalues)

        # Sort singular values in descending order and organize into truncated
        # eigenvector (left-singular vector) and singular value matrices.
        sorted_inds = np.argsort(-s_vals)
        Q = G_eigenvectors[:, sorted_inds]
        s_sorted = s_vals[sorted_inds]
        r = self._compute_rank_from_svals(s_sorted)
        Q = Q[:, :r]
        Sigma = np.diag(s_sorted[:r])

        return Q, Sigma

    def _compute_rank_from_svals(self, s):
        """
        Rank computation for the truncated Singular Value Decomposition given
        only singular values.

        :param numpy.ndarray s: the singular values.
        :return: the computed rank truncation.
        :rtype: int
        """
        if 0 < self._svd_rank < 1:
            cumulative_energy = np.cumsum(s**2 / (s**2).sum())
            return np.searchsorted(cumulative_energy, self._svd_rank) + 1
        elif self._svd_rank >= 1 and isinstance(self._svd_rank, int):
            return min(self._svd_rank, len(s))
        else:
            return len(s)


class EDMD(DMD):
    """
    Kernelized Extended DMD.

    :param svd_rank: the rank for the truncation; If positive integer, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param int tlsq_rank: rank truncation computing Total Least Square. Default
        is 0, meaning no truncation.
    :param opt: If True, amplitudes are computed like in optimized DMD  (see
        :func:`~dmdbase.DMDBase._compute_amplitudes` for reference). If
        False, amplitudes are computed following the standard algorithm. If
        `opt` is an integer, it is used as the (temporal) index of the snapshot
        used to compute DMD modes amplitudes (following the standard
        algorithm).
        The reconstruction will generally be better in time instants near the
        chosen snapshot; however increasing `opt` may lead to wrong results
        when the system presents small eigenvalues. For this reason a manual
        selection of the number of eigenvalues considered for the analyisis may
        be needed (check `svd_rank`). Also setting `svd_rank` to a value
        between 0 and 1 may give better results. Default is False.
    :type opt: bool or int
    :param kernel_metric: the kernel function to apply. Supported kernel
        metrics include `"additive_chi2"`, `"chi2"`, `"linear"`, `"poly"`,
        `"polynomial"`, `"rbf"`, `"laplacian"`, `"sigmoid"`, and `"cosine"`.
    :type kernel_metric: str
    :param kernel_params: additional parameters for the
        `sklearn.metrics.pairwise_kernels` function, including
        kernel-specific function parameters.
    :type kernel_params: dict
    """

    def __init__(
        self,
        svd_rank=-1,
        tlsq_rank=0,
        opt=False,
        kernel_metric="linear",
        kernel_params=None,
    ):
        if svd_rank == 0:
            raise ValueError("svd_rank = 0 functionality is not supported.")

        if kernel_params is None:
            kernel_params = {}
        elif not isinstance(kernel_params, dict):
            raise TypeError("kernel_params must be a dict.")

        self._test_kernel_inputs(kernel_metric, kernel_params)

        super().__init__(
            svd_rank=svd_rank,
            tlsq_rank=tlsq_rank,
            exact=True,
            opt=opt,
        )

        self._Atilde = EDMDOperator(svd_rank, kernel_metric, kernel_params)

        self._kernel_metric = kernel_metric
        self._kernel_params = kernel_params
        self._svd_modes = None

    def eigenfunctions(self, x):
        """
        :param x: array from the original snapshot domain denoting the point
            at which to compute the EDMD eigenfunctions.
        :type x: numpy.ndarray
        :return: array containing all r EDMD eigenfunctions evaluated at the
            given domain point, where r denotes the rank of the EDMD fit.
        :rtype: numpy.ndarray
        """
        if self._svd_modes is None:
            raise ValueError("You need to call fit before")

        if not isinstance(x, np.ndarray) or np.ndim(x) != 1:
            raise ValueError("Input x must be a 1-D numpy array.")

        K_xx = pairwise_kernels(
            x[None],
            self.snapshots.T,
            self._kernel_metric,
            **self._kernel_params,
        )

        return np.linalg.multi_dot(
            [
                K_xx,
                self._svd_modes,
                np.linalg.pinv(np.diag(self.operator.svd_vals)),
                self.operator.eigenvectors,
            ]
        )

    @staticmethod
    def _test_kernel_inputs(kernel_metric, kernel_params):
        """
        Helper function that uses a dummy array of data in order to
        call `sklearn.metrics.pairwise_kernels` using the user-given
        kernel parameters. Ensures that the given kernel parameters
        produce valid kernel matrices.
        """
        x_dummy = np.arange(4).reshape(2, 2)
        try:
            pairwise_kernels(
                x_dummy, x_dummy, metric=kernel_metric, **kernel_params
            )
        except Exception as e:
            msg = (
                "Invalid kernel parameters given to the EDMD model. "
                "Please verify that kernel_metric and kernel_params "
                "are passable to sklearn.metrics.pairwise_kernels."
            )
            raise ValueError(msg) from e



"""
SVD projection pre-processing.
"""

import sys
import warnings
from numbers import Number
from typing import Tuple

import numpy as np

from pydmd.dmdbase import DMDBase
from pydmd.preprocessing.pre_post_processing import (
    PrePostProcessing,
    PrePostProcessingDMD,
)
from pydmd.utils import compute_svd

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


def svd_projection_preprocessing(dmd: DMDBase, svd_rank: Number):
    """
    SVD projection pre-processing.

    :param dmd: DMD instance to be wrapped.
    :param svd_rank: SVD rank argument passed to :func:`pydmd.utils.compute_svd`
        to compute the projection matrix.
    """
    return PrePostProcessingDMD(dmd, _SvdProjectionPrePostProcessing(svd_rank))


class _SvdProjectionPrePostProcessing(PrePostProcessing):
    def __init__(self, svd_rank: Number):
        self._svd_rank = svd_rank

    @override
    def pre_processing(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        space_dim = X.shape[0]
        if space_dim == 1:
            svd_rank = -1
            warnings.warn(
                (
                    f"The parameter {svd_rank=} has "
                    "been ignored because the given system is a scalar function"
                )
            )
        else:
            svd_rank = self._svd_rank
        projection_matrix, _, _ = compute_svd(X, svd_rank)

        return projection_matrix, projection_matrix.T.dot(X)

    @override
    def post_processing(
        self, pre_processing_output: np.ndarray, Y: np.ndarray
    ) -> np.ndarray:
        return pre_processing_output.dot(Y)
